#ifndef BLD_BLOCK_H
#define BLD_BLOCK_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>
#include <wba/wba.h>
#include <bld/op.h>
#include <bld/rule.h>

/***********************************************************************
class BLD_SubBlock
***********************************************************************/
class BLD_SubBlock
{
	public:
		BLD_SubBlock(void) : index(0), shouldIncrement(TRUE) {}
		BLD_SubBlock(const OpArray & ops, const GPL_Array<GPL_String> & exp)
			: index(0), shouldIncrement(TRUE), operations(ops), expressions(exp) {}
		~BLD_SubBlock(void) {}

		GPL_Boolean operator==(const BLD_SubBlock & s) const
		{
			return index == s.index && shouldIncrement == s.shouldIncrement &&
				operations == s.operations && expressions == s.expressions;
		}

		void SetDigitVariables(void);
		void Reset(void);
		GPL_Boolean Next(void);
		void IncludeOperations(const OpArray & ops);
void Dump(void) const;

	protected:
		unsigned long index;
		GPL_Boolean shouldIncrement;
		OpArray operations;
		GPL_Array<GPL_String> expressions;
};

/***********************************************************************
class BLD_Block
***********************************************************************/
class BLD_Block
{
	public:
		BLD_Block(void);
		virtual ~BLD_Block(void) {}

		GPL_Boolean HasPrimarySubBlock(void) const;

		void AddRule(const BLD_RuleBase * rule,
						const GPL_Array<GPL_String> & expressions);
		void IncludeOperations(const OpArray & ops);
		void Terminate(void) { terminatePrematurely = TRUE; }
		void Execute(const GPL_Boolean reuseScope = FALSE);
		void ExecuteFailBlocks(const GPL_Boolean reuseScope = FALSE);

	protected:
		GPL_Boolean terminatePrematurely;
		GPL_List<BLD_SubBlock> subBlocks;
		GPL_List<BLD_SubBlock> failBlocks;
		unsigned long index;
		unsigned long prefix;	// Number of PREPEND sub-blocks
		unsigned long suffix;	// Number of APPEND sub-blocks
		GPL_Boolean   inFail;
};

#endif /* BLD_BLOCK_H */

