#ifndef BLD_FNOP_H
#define BLD_FNOP_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>
#include <bld/op.h>
#include <bld/rule.h>

/***********************************************************************
class BLD_EchoOperation

Print arguments.
***********************************************************************/
class BLD_EchoOperation : public BLD_ListOperation
{
	public:
		BLD_EchoOperation(void) {}
		BLD_EchoOperation(const OpArray & o) : BLD_ListOperation(o) {}
		virtual ~BLD_EchoOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		virtual GPL_Boolean Write(const GPL_String & string) const;
};

/***********************************************************************
class BLD_FileEchoOperation

Print arguments to a file.
***********************************************************************/
class BLD_FileEchoOperation : public BLD_EchoOperation
{
	public:
		BLD_FileEchoOperation(void) : append(FALSE) {}
		BLD_FileEchoOperation(const GPL_Pointer<BLD_Operation> & f,
								const GPL_Boolean a = FALSE)
			: file(f), append(a) {}
		BLD_FileEchoOperation(const OpArray & o,
								const GPL_Pointer<BLD_Operation> & f,
								const GPL_Boolean a = FALSE)
			: BLD_EchoOperation(o), file(f), append(a) {}
		virtual ~BLD_FileEchoOperation(void) {}

		virtual GPL_String ToString(void) const;

	protected:
		GPL_Pointer<BLD_Operation> file;
		GPL_Boolean append;

		virtual GPL_Boolean Write(const GPL_String & string) const;
};

/***********************************************************************
class BLD_SystemOperation

Execute system commands.
***********************************************************************/
class BLD_SystemOperation : public BLD_ListOperation
{
	public:
		BLD_SystemOperation(void) : quiet(FALSE) {}
		BLD_SystemOperation(const OpArray & o, const GPL_Boolean q = FALSE)
			: BLD_ListOperation(o), quiet(q) {}
		virtual ~BLD_SystemOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_Boolean quiet;
};

/***********************************************************************
class BLD_ChangeDirectoryOperation

Change working directory.
***********************************************************************/
class BLD_ChangeDirectoryOperation : public BLD_Operation
{
	public:
		BLD_ChangeDirectoryOperation(BLD_Operation *d) : directory(d) {}
		virtual ~BLD_ChangeDirectoryOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_Pointer<BLD_Operation> directory;
};

/***********************************************************************
class BLD_IncludeOperation

Include file modification times.
***********************************************************************/
class BLD_IncludeOperation : public BLD_Operation
{
	public:
		BLD_IncludeOperation(BLD_Operation *f) : file(f) {}
		virtual ~BLD_IncludeOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_Pointer<BLD_Operation> file;
};

/***********************************************************************
class BLD_ExistOperation

Test file existence.
***********************************************************************/
class BLD_ExistOperation : public BLD_Operation
{
	public:
		BLD_ExistOperation(BLD_Operation *o) : file(o) {}
		virtual ~BLD_ExistOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_Pointer<BLD_Operation> file;
};

/***********************************************************************
class BLD_DateOperation

Get file modification time.
***********************************************************************/
class BLD_DateOperation : public BLD_Operation
{
	public:
		BLD_DateOperation(BLD_Operation *o) : file(o) {}
		virtual ~BLD_DateOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_Pointer<BLD_Operation> file;
};

/***********************************************************************
class BLD_DateSortOperation

Sort file modification times.
***********************************************************************/
class BLD_DateSortOperation : public BLD_ListOperation
{
	public:
		BLD_DateSortOperation(void) {}
		BLD_DateSortOperation(const OpArray & o, const GPL_Boolean a = FALSE)
			: BLD_ListOperation(o) {}
		virtual ~BLD_DateSortOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_Boolean ascending;
};

/***********************************************************************
class BLD_SpecificationOperation

Specify target operations.
***********************************************************************/
class BLD_SpecificationOperation : public BLD_Operation
{
	public:
		BLD_SpecificationOperation(const GPL_String & t,
									const BLD_RuleType y, const OpArray & o)
			: target(t), type(y), ops(o) {}
		virtual ~BLD_SpecificationOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_String target;
		BLD_RuleType type;
		OpArray ops;
};

#endif /* BLD_FNOP_H */
