#!/usr/local/bin/perl -w

require 5.000;
use strict;
use English;
use FileHandle;
require Carp;
require Getopt::Long;
require POSIX;

my($debug) = 0;
my($rc) = 0;
my($date) = POSIX::strftime('%d %B %Y', localtime());
my($package) = undef;
my($module) = 'User Commands';
my($print) = 0;
my($printCommand) = '|troff -man|lp';

sub readFile {
	my($input) = shift;
	unless ($input and open(INPUT, $input)) {
		&Carp::carp("Unable to open \'$input\' for reading");
		return 1;
	}
	unless (!$print or open(PIPE, $printCommand)) {
		&Carp::carp("Unable to open \'$printCommand\' for writing");
		return 1;
	}
	my($item) = undef;
	my($section) = undef;
	my($inMan) = 0;
	my($line);
	while (defined($line = <INPUT>)) {
		if ($line =~ m/END_INLINE_MANPAGE/) {
			$inMan = 0;
			$item = undef;
			$section = undef;
			close(OUTPUT);
		}
		elsif ($line =~ m/BEGIN_INLINE_MANPAGE\s+(\w+)\.(\w+)/) {
			$inMan = 1;
			$item = $::1;
			$section = $::2;
			if ($debug) {
				STDOUT->print("\$item == \'$item\'\n");
				STDOUT->print("\$section == \'$section\'\n");
			}
			unless (open(OUTPUT, ">$item\.$section")) {
				&Carp::carp("Unable to open \'$item\.$section\' for writing");
				return 1;
			}
		}
		elsif ($inMan) {
			$line =~ s/_SECTION_/$section/g;
			$line =~ s/_DATE_/$date/g;
			$line =~ s/_PACKAGE_/$package/g;
			$line =~ s/_MODULE_/$module/g;
			PIPE->print($line) if ($print);
			OUTPUT->print($line);
		}
	}
	close(PIPE) if $print;
	close(INPUT);
	return 0;
}

# Parse the command-line options.
$Getopt::Long::autoabbrev = 1 unless $Getopt::Long::autoabbrev;
&Getopt::Long::GetOptions (
	'debug!' => \$debug,
	'package=s' => \$package,
	'print!' => \$print,
);

unless (@::ARGV && $package) {
	&Carp::carp("Usage: $::PROGRAM_NAME -p package files...");
	exit (1);
}

if ($debug) {
	STDOUT->print("\$date == \'$date\'\n");
	STDOUT->print("\$package == \'$package\'\n");
	STDOUT->print("\$module == \'$module\'\n");
}

foreach $::ARG (@::ARGV) {
	$rc += &readFile($::ARG);
}

exit ($rc);

__END__
